#include "Distributor.h"
#include <boost/shared_ptr.hpp>
#include <boost/bind.hpp>
#include <iostream>
#include <sys/time.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "Reader.h"
#include "WFCounter.h"

Distributor::Distributor( const char * filename, int numReader, int numWorker )
: m_filename( filename )
, m_numReader( numReader )
, m_numWorker( numWorker )
, m_finished( 0 )
, m_finishedReaders( 0 )
{
    struct stat64 s;
    stat64( m_filename, &s );

    off64_t size = s.st_size;
    off64_t start = 0;
    off64_t step = ( size + m_numReader - 1 ) / m_numReader;

    for ( int i = 0; i < m_numReader; ++i ) {
        Reader * reader( new Reader( this, m_filename, start, start + step - 1 ) );
        m_threads.create_thread( boost::bind( &Reader::run, reader ) );
        m_readers.push_back( reader );
        start += step;
    }

    for (int i = 0; i < m_numWorker; ++i ) {
        WFCounter * wc( new WFCounter( this ) );
        m_threads.create_thread( boost::bind( &WFCounter::run, wc ) );
        m_counters.push_back( wc );
    }
}

Distributor::~Distributor()
{
    m_threads.join_all();
}

int Distributor::process()
{
    boost::mutex::scoped_lock lock( m_finishedMutex );
    if ( m_finishedReaders != m_numReader ) {
        m_finishedCondition.wait( lock );
    }
    return 0;
}

void Distributor::deliverResult( Result * result[Last] )
{
    for ( int i = 0; i < Last; ++i ) {
        deliverResult( result[i] );
    }

    while ( m_finished != Last ) {
        if ( ! processNext() ) {
            usleep( 10000 );
        }
    }
}

void Distributor::deliverResult( Result * result )
{
    boost::mutex::scoped_lock lock( m_resultMutex );
    if ( result->count == m_numWorker ) {
        m_done[result->type] = result;
        ++m_finished;
        if ( m_finished == Last ) {
            printf( "%d resources, %d 404s, %d clients\n\n", m_done[Hits]->data.size(), m_done[FourOFours]->data.size(), m_done[Clients]->data.size() );
            Type order[] = { Hits, Bytes, FourOFours, Clients, Refs };
            for ( unsigned int i = 0; i < sizeof( order ) / sizeof( Type ); ++i )  {
                m_done[i]->report();
            }
            m_finishedCondition.notify_one();
        }
    }
    else {
        m_results[result->type].push_back( result );
    }
}

bool Distributor::processNext()
{
    boost::mutex::scoped_lock lock( m_resultMutex );
    for ( int i = 0; i < Last; ++i ) {
        if ( m_results[i].size() >= 2  ) {
            Result * first = m_results[i].front();
            m_results[i].pop_front();
            Result * second = m_results[i].front();
            m_results[i].pop_front();
            lock.unlock();
            first->merge( second );
            deliverResult( first );
            return true;
        }
    }
    return false;
}

void Distributor::readerFinished( Reader * reader )
{
    boost::mutex::scoped_lock lock( m_finishedMutex );
    ++m_finishedReaders;
    if ( m_finishedReaders == m_numReader ) {
        for ( int i = 0; i < m_numWorker; ++i ) {
            pushBuffer( 0 );
        }
    }
}

void Distributor::pushBuffer( Buffer * buffer )
{
    boost::mutex::scoped_lock lock( m_buffersMutex );
    m_buffers.push_back( buffer );
    m_bufferAvailable.notify_one();
}

Buffer * Distributor::popBuffer()
{
    boost::mutex::scoped_lock lock( m_buffersMutex );
    while ( m_buffers.empty() ) {
        m_bufferAvailable.wait(  lock );
    }

    Buffer * buffer = m_buffers.front();
    m_buffers.pop_front();
    return buffer;
}
